def test_binary_tree():
    # wikipedia example, https://en.wikipedia.org/wiki/Binary_tree#Arrays
    
    # root test
    assert bt_is_root(0)
    assert bt_left_child(0) == 1
    assert bt_right_child(0) == 2
    assert bt_node_from_left(0, 0) == 0
    assert bt_offset_from_left(0) == 0 
    assert bt_offset_from_right(0) == 0 
    assert bt_left_leaf(0, 3) == 7
    assert bt_right_leaf(0, 3) == 14

    # lvl 1
    assert bt_is_left_child(1)
    assert bt_is_right_child(2)
    assert bt_left_child(1) == 3
    assert bt_right_child(1) == 4
    assert bt_left_child(2) == 5
    assert bt_right_child(2) == 6
    assert bt_parent(1) == 0
    assert bt_parent(2) == 0
    assert bt_right_sibling(1) == 2
    assert bt_left_sibling(2) == 1
    assert bt_node_from_left(0, 1) == 1
    assert bt_node_from_left(1, 1) == 2
    assert bt_offset_from_left(1) == 0 
    assert bt_offset_from_left(2) == 1 
    assert bt_offset_from_right(1) == 1 
    assert bt_offset_from_right(2) == 0 
    assert bt_left_leaf(1, 3) == 7
    assert bt_left_leaf(2, 3) == 11
    assert bt_right_leaf(1, 3) == 10
    assert bt_right_leaf(2, 3) == 14
    
    # lvl 2
    assert bt_is_left_child(3)
    assert bt_is_right_child(4)
    assert bt_is_left_child(5)
    assert bt_is_right_child(6)
    assert bt_parent(3) == 1
    assert bt_parent(4) == 1
    assert bt_parent(5) == 2
    assert bt_parent(6) == 2
    assert bt_right_sibling(3) == 4
    assert bt_left_sibling(4) == 3
    assert bt_right_sibling(5) == 6
    assert bt_left_sibling(6) == 5
    assert bt_node_from_left(0, 2) == 3
    assert bt_node_from_left(1, 2) == 4
    assert bt_node_from_left(2, 2) == 5
    assert bt_node_from_left(3, 2) == 6
    assert bt_offset_from_left(3) == 0 
    assert bt_offset_from_left(4) == 1 
    assert bt_offset_from_left(5) == 2 
    assert bt_offset_from_left(6) == 3
    assert bt_offset_from_right(3) == 3
    assert bt_offset_from_right(4) == 2 
    assert bt_offset_from_right(5) == 1 
    assert bt_offset_from_right(6) == 0
    assert bt_left_leaf(3, 3) == 7
    assert bt_left_leaf(4, 3) == 9
    assert bt_left_leaf(5, 3) == 11
    assert bt_left_leaf(6, 3) == 13
    assert bt_right_leaf(3, 3) == 8
    assert bt_right_leaf(4, 3) == 10
    assert bt_right_leaf(5, 3) == 12
    assert bt_right_leaf(6, 3) == 14

    # lvl 3
    assert bt_offset_from_left(7) == 0
    assert bt_offset_from_left(8) == 1
    assert bt_offset_from_left(9) == 2
    assert bt_offset_from_left(10) == 3
    assert bt_offset_from_left(11) == 4
    assert bt_offset_from_left(12) == 5
    assert bt_offset_from_left(13) == 6
    assert bt_offset_from_left(14) == 7
    assert bt_offset_from_right(7) == 7
    assert bt_offset_from_right(8) == 6
    assert bt_offset_from_right(9) == 5
    assert bt_offset_from_right(10) == 4
    assert bt_offset_from_right(11) == 3
    assert bt_offset_from_right(12) == 2
    assert bt_offset_from_right(13) == 1
    assert bt_offset_from_right(14) == 0
    



cdef extern from "<inttypes.h>":
    ctypedef unsigned int uint64_t

cdef extern from "bit_array.h":
    struct BIT_ARRAY:
        pass
    ctypedef uint64_t bit_index_t

    # allocations
    BIT_ARRAY* bit_array_create(bit_index_t nbits)
    void bit_array_free(BIT_ARRAY* bitarray)
    BIT_ARRAY* bit_array_clone(const BIT_ARRAY* bitarr)
   
    # utility
    char* bit_array_to_str(const BIT_ARRAY* bitarr, char* str)
    bit_index_t bit_array_length(const BIT_ARRAY* bit_arr)
    bit_index_t bit_array_num_bits_set(const BIT_ARRAY* bitarr)

    # bit juggling
    void bit_array_set_bit(BIT_ARRAY* bitarr, bit_index_t b) nogil
    void bit_array_toggle_bit(BIT_ARRAY* bitarr, bit_index_t b) nogil
    char bit_array_get_bit(const BIT_ARRAY* bitarr, bit_index_t b) nogil
    void bit_array_clear_bit(BIT_ARRAY* bitarr, bit_index_t b)

    # logical operations
    void bit_array_and(BIT_ARRAY* dest, const BIT_ARRAY* src1, const BIT_ARRAY* src2)
    void bit_array_or(BIT_ARRAY* dest, const BIT_ARRAY* src1, const BIT_ARRAY* src2)
    void bit_array_xor(BIT_ARRAY* dest, const BIT_ARRAY* src1, const BIT_ARRAY* src2)
    void bit_array_not(BIT_ARRAY* dest, const BIT_ARRAY* src1, const BIT_ARRAY* src2)

    # cyclic shifting
    void bit_array_cycle_right(BIT_ARRAY* bitarr, bit_index_t dist)
    void bit_array_cycle_left (BIT_ARRAY* bitarr, bit_index_t dist)



cimport libc.stdlib
from cpython cimport Py_INCREF, Py_DECREF, bool
cimport numpy as np
np.import_array()

cdef class bitarray:
    cdef:
        BIT_ARRAY* bitarr
        bit_index_t nbits   ### REDUNDANT as the BIT_ARRAY struct has this

    def __cinit__(self, bit_index_t n):
        self.nbits = n
        self.bitarr = bit_array_create(n)

    def __dealloc__(self):
        bit_array_free(self.bitarr)

    def __str__(self):
        cdef char* str_
        cdef object result

        str_ = <char*>libc.stdlib.malloc(self.nbits + 1)
        result = tounicode(bit_array_to_str(self.bitarr, str_))
        libc.stdlib.free(str_)

        return result

    def __getitem__(self, bit_index_t i):
        # could inline the macro version as well, less safe but should be 
        # thinner code
        return bit_array_get_bit(self.bitarr, i)

    def __setitem__(self, bit_index_t i, bool v):
        if v:
            bit_array_set_bit(self.bitarr, i)
        else:
            bit_array_clear_bit(self.bitarr, i)

    

cdef unicode tounicode(char* s):
    # from http://docs.cython.org/en/latest/src/tutorial/strings.html
    return s.decode('UTF-8', 'strict')


cpdef bitarray bitarray_factory(object vec):
    """Construct a bitarray from a vector

    A value at a position in the vector is considered set if the value 
    evaluates True.

    Parameters
    ----------
    vec : object
        An iterable

    Returns
    -------
    bitarray
        The bitarray based off of topo
    """
    cdef int i
    cdef int n = len(vec)
    cdef bitarray result
    
    result = bitarray(n)
    for i in range(n):
        if vec[i]:
            result[i] = True

    return result



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

# An implementation of a complete binary tree in breadth first order adapted 
# from https://github.com/jfuentess/sea2015/blob/master/binary_trees.h

from libc.math cimport pow, log2, floor
from bp._bp cimport SIZE_t


# it might be useful to use a pow2 lookup. static const c arrays are not 
# allowed, so it might be useful to do it as a memoryview but should only be
# done following benching

cdef inline SIZE_t bt_is_root(SIZE_t v) nogil:
    """Is v the root"""
    return v == 0


cdef inline SIZE_t bt_is_left_child(SIZE_t v) nogil:
    """Is v a left child of some node"""
    return 0 if bt_is_root(v) else v % 2


cdef inline SIZE_t bt_is_right_child(SIZE_t v) nogil:
    """Is v a right child of some node"""
    return 0 if bt_is_root(v) else 1 - (v % 2)


cdef inline SIZE_t bt_parent(SIZE_t v) nogil:
    """Get the index of the parent of v"""
    return 0 if bt_is_root(v) else (v - 1) // 2


cdef inline SIZE_t bt_left_child(SIZE_t v) nogil:
    """Get the index of the left child of v"""
    return 2 * v + 1


cdef inline SIZE_t bt_right_child(SIZE_t v) nogil:
    """Get the index of the right child of v"""
    return 2 * v + 2


cdef inline SIZE_t bt_left_sibling(SIZE_t v) nogil:
    """Get the index of the left sibling of v"""
    return v - 1


cdef inline SIZE_t bt_right_sibling(SIZE_t v) nogil:
    """Get the index of the right sibling of v"""
    return v + 1


cdef inline SIZE_t bt_is_leaf(SIZE_t v, SIZE_t height) nogil:
    """Determine if v is a leaf"""
    return <SIZE_t>(v >= pow(2, height) - 1)


cdef inline SIZE_t bt_node_from_left(SIZE_t pos, SIZE_t height) nogil:
    """Get the index from the left of a node at a given height"""
    return <SIZE_t>pow(2, height) - 1 + pos


cdef inline SIZE_t bt_offset_from_left(SIZE_t v) nogil:
    """Get the position from left of a node at its level

    This is the inverse of bt_node_from_left
    """
    cdef double leftmost_check

    if bt_is_root(v):
        return 0

    leftmost_check = log2(v + 1)
    if leftmost_check == floor(leftmost_check):
        return 0

    return v - <SIZE_t>pow(2, floor(log2(v))) + 1


cdef inline SIZE_t bt_offset_from_right(SIZE_t v) nogil:
    """Get the position from right of a node at its level"""
    cdef SIZE_t lvl = <SIZE_t>floor(log2(v + 1))
    cdef SIZE_t n_nodes_at_lvl = <SIZE_t>pow(2, lvl)

    return n_nodes_at_lvl - bt_offset_from_left(v) - 1


cdef inline SIZE_t bt_left_leaf(SIZE_t v, SIZE_t height) nogil:
    """Determine the index of a nodes left most leaf"""
    cdef SIZE_t left_tip = <SIZE_t>pow(2, height) - 1
    cdef SIZE_t block_size

    if bt_is_root(v):
        return left_tip

    block_size = <SIZE_t>pow(2, height - floor(log2(v + 1)))

    return left_tip + (block_size * bt_offset_from_left(v))


cdef inline SIZE_t bt_right_leaf(SIZE_t v, SIZE_t height) nogil:
    """Determine the index of a nodes right most leaf"""
    cdef SIZE_t right_tip = <SIZE_t>pow(2, height + 1) - 2
    cdef SIZE_t block_size

    if bt_is_root(v):
        return right_tip

    block_size = <SIZE_t>pow(2, height - floor(log2(v + 1)))

    return right_tip - (block_size * bt_offset_from_right(v))





# ----------------------------------------------------------------------------
# Copyright (c) 2013--, BP development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
from ._bp import BP
from ._io import parse_newick, write_newick, parse_jplace
from ._conv import to_skbio_treenode, from_skbio_treenode, to_skbio_treearray
from ._insert import insert_fully_resolved


def load(fp, as_treenode=False):
    data = open(fp).read()
    tree = parse_newick(data)
    if as_treenode:
        return to_skbio_treenode(tree)
    else:
        return tree


__all__ = ['BP', 'parse_newick', 'to_skbio_treenode', 'from_skbio_treenode',
           'to_skbio_treearray', 'write_newick', 'parse_jplace',
           'insert_fully_resolved', 'load']

from . import _version
__version__ = _version.get_versions()['version']



import unittest
import pkg_resources
from bp import parse_jplace, insert_fully_resolved
import skbio
import pandas as pd


class InsertTests(unittest.TestCase):
    package = 'bp.tests'
    def setUp(self):
        self.jplacedata_multiple = \
            open(self.get_data_path('300/placement_mul.jplace')).read()
        self.final_multiple_fully_resolved = \
            skbio.TreeNode.read(self.get_data_path('300/placement.full_resolve.newick'))

    def get_data_path(self, filename):
        # adapted from qiime2.plugin.testing.TestPluginBase
        return pkg_resources.resource_filename(self.package,
                                               'data/%s' % filename)

    def test_insert_fully_resolved(self):
        exp = self.final_multiple_fully_resolved
        placements, backbone = parse_jplace(self.jplacedata_multiple)
        obs = insert_fully_resolved(placements, backbone)
        self.assertEqual({n.name for n in obs.tips()},
                         {n.name for n in exp.tips()})
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_tip_distances(exp), 0)

    def test_insert_fully_resolved_multiple_placements(self):
        exp = self.final_multiple_fully_resolved
        placements, backbone = parse_jplace(self.jplacedata_multiple)

        # add another placement elsewhere that we would not keep
        # as it's ratio is lower
        dup1 = placements.iloc[0].copy()
        dup1['like_weight_ratio'] -= 0.5
        dup1['edge_num'] += 1

        # add another placement elsewhere that we would not keep
        # as, though its ratio is the same, its pendant is larger
        dup2 = placements.iloc[1].copy()
        dup2['pendant_length'] += 0.5
        dup2['edge_num'] += 1

        placements = pd.concat([placements, pd.DataFrame([dup1, dup2])])

        obs = insert_fully_resolved(placements, backbone)
        self.assertEqual({n.name for n in obs.tips()},
                         {n.name for n in exp.tips()})
        self.assertEqual(obs.compare_rfd(exp), 0)
        self.assertAlmostEqual(obs.compare_tip_distances(exp), 0)

if __name__ == '__main__':
    unittest.main()



cimport numpy as np
cimport cython

from bp._ba cimport BIT_ARRAY

ctypedef np.npy_intp SIZE_t
ctypedef np.npy_uint32 UINT32_t
ctypedef np.npy_int32 INT32_t
ctypedef np.npy_float64 DOUBLE_t
ctypedef np.npy_uint8 BOOL_t


cdef class mM:
    cdef int b  # block size
    cdef int n_tip  # number of tips in the binary tree
    cdef int n_internal  # number of internal nodes in the binary tree
    cdef int n_total  # total number of nodes in the binary tree
    cdef int height  # the height of the binary tree
    cdef int m_idx  # m is minimum excess
    cdef int M_idx  # M is maximum excess
    cdef int r_idx  # rank
    cdef SIZE_t[:, ::1] mM
    cdef SIZE_t[:] r

    cdef void rmm(self, BOOL_t[:] B, int B_size) nogil


@cython.final
cdef class BP:
    cdef:
        public np.ndarray B 
        BOOL_t* _b_ptr
        SIZE_t[:] _e_index
        SIZE_t[:] _k_index_0
        SIZE_t[:] _k_index_1
        np.ndarray _names
        np.ndarray _lengths
        np.ndarray _edges
        np.ndarray _edge_lookup
        mM _rmm
        SIZE_t size

    cdef inline SIZE_t rank(self, SIZE_t t, SIZE_t i) nogil
    cdef inline SIZE_t select(self, SIZE_t t, SIZE_t k) nogil
    cdef SIZE_t _excess(self, SIZE_t i) nogil
    cdef SIZE_t excess(self, SIZE_t i) nogil
    cdef SIZE_t fwdsearch(self, SIZE_t i, int d) nogil
    cdef SIZE_t bwdsearch(self, SIZE_t i, int d) nogil
    cpdef inline SIZE_t close(self, SIZE_t i) nogil
    cdef inline SIZE_t open(self, SIZE_t i) nogil
    cpdef inline BOOL_t isleaf(self, SIZE_t i) nogil
    cdef inline SIZE_t enclose(self, SIZE_t i) nogil
    cdef BP _mask_from_self(self, BIT_ARRAY* mask, np.ndarray[DOUBLE_t, ndim=1] lengths)
    cpdef SIZE_t nsibling(self, SIZE_t i) nogil
    cpdef SIZE_t psibling(self, SIZE_t i) nogil
    cpdef SIZE_t lchild(self, SIZE_t i) nogil
    cpdef SIZE_t fchild(self, SIZE_t i) nogil
    cpdef SIZE_t parent(self, SIZE_t i) nogil
    cpdef SIZE_t depth(self, SIZE_t i) nogil
    cpdef SIZE_t root(self) nogil
    cdef int scan_block_forward(self, int i, int k, int b, int d) nogil
    cdef int scan_block_backward(self, int i, int k, int b, int d) nogil
    cdef void _set_edges(self, np.ndarray[INT32_t, ndim=1] edges)

    # TODO: evalute down the road what methods should be cdef. There is a 
    # performance cost for cpdef, so for high use functions, it may make sense
    # to punt down to cdef.
    # http://notes-on-cython.readthedocs.io/en/latest/fibo_speed.html
    cpdef inline unicode name(self, SIZE_t i)
    cpdef inline DOUBLE_t length(self, SIZE_t i)
    cpdef inline INT32_t edge(self, SIZE_t i)
    cpdef SIZE_t edge_from_number(self, INT32_t n)
    cpdef SIZE_t rmq(self, SIZE_t i, SIZE_t j) nogil
    cpdef SIZE_t rMq(self, SIZE_t i, SIZE_t j) nogil
    cpdef SIZE_t postorderselect(self, SIZE_t k) nogil
    cpdef SIZE_t postorder(self, SIZE_t i) nogil
    cpdef SIZE_t preorderselect(self, SIZE_t k) nogil
    cpdef SIZE_t preorder(self, SIZE_t i) nogil
    cpdef BOOL_t isancestor(self, SIZE_t i, SIZE_t j) nogil
    cpdef SIZE_t levelancestor(self, SIZE_t i, SIZE_t d) nogil
    cpdef SIZE_t subtree(self, SIZE_t i) nogil
    cpdef BP shear(self, set tips)
    cpdef BP collapse(self)
    cpdef SIZE_t ntips(self) nogil
    cpdef SIZE_t levelnext(self, SIZE_t i) nogil
    cpdef SIZE_t height(self, SIZE_t i) nogil
    cpdef SIZE_t deepestnode(self, SIZE_t i) nogil
    cpdef SIZE_t lca(self, SIZE_t i, SIZE_t j) nogil



# encoding: utf-8
# cython: profile=False, boundscheck=False, wraparound=False

from ._bp cimport BP
import time
import numpy as np
import pandas as pd
import json
cimport numpy as np
cimport cython
np.import_array()


cdef inline np.double_t length_from_edge(unicode token):
    cdef:
        Py_ssize_t split_idx

    # 0.12345{0123} -> 0.12345
    # OR 0.12345[0123] -> 0.12345
    split_idx_curly = token.find('{')
    split_idx_square = token.find('[')
    split_idx = max(split_idx_curly, split_idx_square)
    if split_idx == -1:
        return np.double(token)
    else:
        return np.double(token[:split_idx])


cdef inline np.int32_t number_from_edge(unicode token):
    cdef:
        Py_ssize_t split_idx
        Py_ssize_t end

    # 0.12345{0123} -> 0123
    # OR 0.12345[0123] -> 0123
    split_idx_curly = token.find('{')
    split_idx_square = token.find('[')
    split_idx = max(split_idx_curly, split_idx_square)
    if split_idx == -1:
        return 0
    else:
        end = len(token)
        return np.int32(token[split_idx + 1:end - 1])


cdef void _set_node_metadata(np.uint32_t ptr, unicode token,
                             np.ndarray[object, ndim=1] names, 
                             np.ndarray[np.double_t, ndim=1] lengths,
                             np.ndarray[np.int32_t, ndim=1] edges):
    """Inplace update of names and lengths given token details"""
    cdef:
        np.double_t length
        np.int32_t edge
        Py_ssize_t split_idx, i, end
        unicode name, token_parsed

    name = None
    length = 0.0
    edge = 0

    # NOTE: there is likely some fat to trim in this method. we do a lot
    # of work per token, we could probably do that work smarter. as is,
    # the changes to support edge numbers increase parsing ~20%, which
    # is annoying but probably not a critical
    if token[0] == u':':
        token_parsed = token[1:]
        length = length_from_edge(token_parsed)
        edge = number_from_edge(token_parsed)
    elif u':' in token:
        split_idx = token.rfind(':')
        name = token[:split_idx]
        token_parsed = token[split_idx + 1:]
        length = length_from_edge(token_parsed)
        edge = number_from_edge(token_parsed)
        name = name.strip("'").strip()
    elif u'{' in token or u'[' in token:
        # strip as " {123}" is valid?
        token = token.strip()
        end = len(token)
        edge = np.int32(token.strip()[1:end - 1])
    else:
        name = token.replace("'", "").replace('"', "").strip()

    names[ptr] = name
    lengths[ptr] = length
    edges[ptr] = edge


def write_newick(BP tree, object output, bint include_edge):
    cdef:
        list name_stack
        list edge_stack
        list length_stack
        list open_paren_stack
        object name
        np.npy_float64 length
        Py_ssize_t idx
        np.npy_uint8 v
        Py_ssize_t root_close

    length_stack = []
    name_stack = []
    edge_stack = []
    open_paren_stack = []
    root_close = tree.close(0)

    for idx, v in enumerate(tree.B):
        if v:
            if not tree.isleaf(idx):
                output.write('(')
            name_stack.append(tree.name(idx))
            length_stack.append(tree.length(idx))
            edge_stack.append(tree.edge(idx))
            open_paren_stack.append(idx)
        else:
            name = name_stack.pop()
            length = length_stack.pop()
            edge = edge_stack.pop()

            if name is not None:
                # if we have magical characters, make sure we quote
                if set(name) & {';', ',', '(', ')', ':', '_'}:
                    output.write("'%s'" % name)
                else:
                    output.write(name)

            if include_edge:
                output.write(':%f{%d}' % (length, edge))
            else:
                output.write(':%f' % length)
                
            if tree.nsibling(open_paren_stack.pop()) == 0:
                if idx != root_close:
                    output.write(')')
            else:
                output.write(',')

    output.write(';')


cpdef parse_newick(unicode data):
    cdef:
        np.uint32_t ptr, open_ptr
        Py_ssize_t token_ptr, tmp, lag, datalen
        BP topology
        unicode token, last_token
        np.ndarray[object, ndim=1] names
        np.ndarray[np.double_t, ndim=1] lengths
        np.ndarray[np.int32_t, ndim=1] edges


    data = data.strip()
    if not data.endswith(';'):
        raise ValueError("Newick does not appear terminated with a semicolon")

    datalen = len(data)
    topology = _newick_to_bp(data)
    
    if len(topology.B) <= 2:
        raise ValueError("Only trees with more than 1 node supported")

    names = np.full(len(topology.B), None, dtype=object)
    lengths = np.zeros(len(topology.B), dtype=np.double)
    edges = np.full(len(topology.B), 0, dtype=np.int32) 

    ptr = 0
    token_ptr = _ctoken(data, datalen, 0)
    token = data[0:token_ptr]
    last_token = None

    # lag reflects the scenario where ((x))y, where the label y gets may end
    # up being associated with an earlier unnamed vertex. lag represents the
    # offset between the topology pointer and the token pointer effectively.
    lag = 0
    while token != ';':
        if token == '(':
            # an open parenthesis never has metadata associated with it
            ptr += 1

        if (token == ')' or token == ',') and last_token == ')':
            # determine if there are unnamed/unlengthed nodes 
            lag += 1

        elif token not in '(),:;':
            ptr += lag
            lag = 0

            open_ptr = topology.open(ptr)
            _set_node_metadata(open_ptr, token, names, lengths, edges)

            if topology.isleaf(ptr):
                ptr += 2
            else:
                ptr += 1

        last_token = token
        tmp = _ctoken(data, datalen, token_ptr)
        token = data[token_ptr:tmp]
        token_ptr = tmp

    topology.set_names(names)
    topology.set_lengths(lengths)
    topology.set_edges(edges)

    return topology


cdef object _newick_to_bp(unicode data):
    """Convert newick to balanced parentheses

    Newick is _similar_ to BP, but differs notably at the tips of the tree.
    The complexity of the parse below comes from handling tips, and single
    descendents. Examples of situations that introduce this complexity are:

    ((a,b)) -> 11101000
    (a) -> 1100
    () -> 1100
    ((a,b),c) -> 1110100100
    (a,(b,c)) -> 1101101000

    Newick is not required to have node labels on tips, and the interpretation
    of a comma is dependent on prior state.

    The strategy undertaken is to reduce the newick string to only structural
    components. From there, the string is interpreted into tokens of: {"1",
    "0", "10", "100"}, which directly translate into the resulting balanced
    parentheses topology.

    It is very likely the case that this parser can be done better with
    improved efficiency.
    """
    cdef:
        Py_ssize_t i, topology_ptr, single_descendent
        Py_UCS4 c, last_c
        np.ndarray[np.uint8_t, ndim=1] topology
        
    potential_single_descendant = False

    topology = np.empty(len(data), dtype=np.uint8)
    topology_ptr = 0
    last_c = u'x'
    in_quote = False

    for i in range(len(data)):
        c = data[i]
        if c == u"'":
            in_quote = not in_quote
        else:
            if in_quote:
                continue
            elif c == u'(':
                # opening of a node
                topology[topology_ptr] = 1
                topology_ptr += 1
                last_c = c
                potential_single_descendant = True
            elif c == u')':
                # closing of a node
                if potential_single_descendant or last_c == u',':
                    # we have a single descendant or a last child (i.e., ",)")
                    topology[topology_ptr] = 1
                    topology[topology_ptr + 1] = 0
                    topology[topology_ptr + 2] = 0
                    topology_ptr += 3
                    potential_single_descendant = False
                else:
                    # it is possible to still have a single descendant in the case
                    # of a multiple single descendant: (...()...)
                    topology[topology_ptr] = 0
                    topology_ptr += 1
                last_c = c
            elif c == u',':
                if last_c != u')':
                    # we have a new tip
                    topology[topology_ptr] = 1
                    topology[topology_ptr + 1] = 0
                    topology_ptr += 2
                potential_single_descendant = False
                last_c = c
            else:
                # ignore non-structure
                pass

    return BP(topology[:topology_ptr])


cdef inline int _ccheck(Py_UCS4 c):
    """structure check"""
    cdef:
        Py_ssize_t i

    if c == u'(':
        return 1
    elif c == u')':
        return 1
    elif c == u',':
        return 1
    elif c == u';':
        return 1
    else:
        return 0


cdef inline int _is_quote(Py_UCS4 c):
    if c == u'"':
        return 1
    elif c == u"'":
        return 1
    else:
        return 0


cdef inline Py_ssize_t _ctoken(unicode data, Py_ssize_t datalen, Py_ssize_t start):
    cdef:
        Py_ssize_t idx, in_quote = 0
        Py_UCS4 c

    if start == (datalen - 1):
        return start + 1

    for idx in range(start, datalen):
        c = data[idx]

        if in_quote:
            if _is_quote(c):
                in_quote = 0
            continue
        else:
            if _is_quote(c):
                in_quote = 1
                continue

        if _ccheck(c):
            if idx == start:
                return idx + 1
            else:
                return idx
    
    return idx + 1


def parse_jplace(object data):
    """Takes a jplace string, returns a DataFrame of placements and the tree

    Implementation specific caveats:

    1) we do not support multiplicities. placements are required to have an "n"
        entry, and we ignore "nm"
    2) Matsen et al (https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0031009) 
        define [] for denoting edge labels and {} for denoting edge numbers. We
        currently support either [] OR {}, we do not support edges with both.
        In addition, we REQUIRE the edge labels if specified to be integer.

    If either of these caveats are problems, then we need to modify the code.
    """
    cdef:
        dict as_json
        list fields, placements, fragments, p, placement_data, 
        list placement_inner_data, pquery, entry
        unicode frag, newick
        Py_ssize_t placement_idx, placement_inner_idx, fragment_idx, 
        Py_ssize_t n_fragments
        BP tree
        object df
        set edges

    as_json = json.loads(data)
    newick = as_json['tree']
    placement_data = as_json['placements']

    fields = as_json['fields']
    fields = ['fragment', ] + fields

    placements = []
    for placement_idx in range(len(placement_data)):
        placement = placement_data[placement_idx]
        
        placement_inner_data = placement['p']

        if 'n' not in placement:
            raise KeyError("jplace parsing limited to entries with 'n' keys")

        fragments = placement['n']
        n_fragments = len(fragments)

        for placement_inner_idx in range(len(placement_inner_data)):
            pquery = placement_inner_data[placement_inner_idx]

            for fragment_idx in range(n_fragments):
                frag = fragments[fragment_idx]
                entry = [frag, ] + pquery
                placements.append(entry)

    tree = parse_newick(newick)
    edges = {tree.edge(i) for i, v in enumerate(tree.B) if v}
    df = pd.DataFrame(placements, columns=fields)
    df = df[df['edge_num'].isin(edges)]

    return df, tree



from unittest import TestCase, main
from bp import parse_newick, to_skbio_treenode, write_newick, parse_jplace
import re
import json
import pandas as pd
import pandas.testing as pdt
import pkg_resources
import skbio
import io
import numpy as np
import numpy.testing as npt


class NewickTests(TestCase):
    def test_parse_newick_simple_edge_numbers(self):
        # from https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0031009
        # but without edge labels
        # 0 1 2 3 4 5 6 7 8 9
        # 1 1 1 0 1 0 0 1 0 0
        in_ = '((A:.01{0}, B:.01{1})D:.01{3}, C:.01{4}) {5};'
        exp_sk = '((A:.01, B:.01)D:.01, C:.01);'  # skbio doesn't know about edge numbers
        obs = parse_newick(in_)
        obs_sk = to_skbio_treenode(obs)
        exp_sk = skbio.TreeNode.read([exp_sk])
        self.assertEqual(obs_sk.compare_rfd(exp_sk), 0)

        self.assertEqual(obs.edge(2), 0)
        self.assertEqual(obs.edge(4), 1)
        self.assertEqual(obs.edge(1), 3)
        self.assertEqual(obs.edge(7), 4)
        self.assertEqual(obs.edge(0), 5)
        self.assertEqual(obs.edge_from_number(0), 2)
        self.assertEqual(obs.edge_from_number(1), 4)
        self.assertEqual(obs.edge_from_number(3), 1)
        self.assertEqual(obs.edge_from_number(4), 7)
        self.assertEqual(obs.edge_from_number(5), 0)

    def _compare_newick(self, obs, exp):
        a = skbio.TreeNode.read([obs])
        b = skbio.TreeNode.read([exp])
        self.assertEqual(a.compare_rfd(b), 0)
        npt.assert_equal(a.tip_tip_distances().data,
                         b.tip_tip_distances().data)

    def test_write_newick_cases(self):
        tests = ['((foo"bar":1,baz:2)x:3)r;',
                 "(((a:1,b:2.5)c:6,d:8,(e),(f,g,(h:1,i:2)j:1)k:1.2)l,m:2)r;",
                 "(((a)b)c,((d)e)f)r;",
                 "((a,(b,c):5)'d','e; foo':10,((f))g)r;"]

        for test in tests:
            buf = io.StringIO()
            obs = write_newick(parse_newick(test), buf, False)
            buf.seek(0)
            obs = buf.read()
            self._compare_newick(obs, test)

    def test_write_newick_edges(self):
        test_a = '((foo"bar":1{0},baz:2{1})x:3{2})r;'
        test_b = "(((a)b)c,((d)e)f)r;"

        buf = io.StringIO()
        obs = write_newick(parse_newick(test_a), buf, True)
        buf.seek(0)
        obs = to_skbio_treenode(parse_newick(buf.read()))
        self.assertEqual(obs.find('foo"bar"').edge_num, 0)
        self.assertEqual(obs.find('baz').edge_num, 1)
        self.assertEqual(obs.find('x').edge_num, 2)

        buf = io.StringIO()
        obs = write_newick(parse_newick(test_b), buf, True)
        buf.seek(0)
        obs = to_skbio_treenode(parse_newick(buf.read()))
        for o in obs.traverse():
            self.assertEqual(o.edge_num, 0)

    def test_parse_newick_singlenode_bug(self):
        # https://github.com/wasade/improved-octo-waddle/issues/29
        test = 'i:1;'

        # let's not allow this edge case
        with self.assertRaises(ValueError):
            parse_newick(test)

    def test_parse_newick_no_semicolon_bug(self):
        # https://github.com/wasade/improved-octo-waddle/issues/26
        test = "((h:1, i:1, j:1, k:1, l: 1),(e:1,f:1),(n:1,o:1,p:1))a:1"

        with self.assertRaises(ValueError):
            parse_newick(test)

        # make sure we work with a newline
        test = "((h:1, i:1, j:1, k:1, l: 1),(e:1,f:1),(n:1,o:1,p:1))a:1;\n"
        parse_newick(test)

    def test_write_newick_underscore_bug(self):
        test = "(((a)b)'c_foo',((d)e)f)r;"
        buf = io.StringIO()
        obs = write_newick(parse_newick(test), buf, False)
        buf.seek(0)
        self.assertIn("'c_foo'", test)

    def test_parse_newick_nested_quotes(self):
        # bug: quotes are removed
        in_ = '((foo"bar":1,baz:2)x:3)r;'
        exp = skbio.TreeNode.read([in_])
        obs = to_skbio_treenode(parse_newick(in_))
        self.assertEqual(obs.compare_rfd(exp), 0.0)

    def test_parse_newick_with_commas(self):
        # bug: comma is getting interpreted even if in quotes
        in_ = "(('foo,bar':1,baz:2)x:3)r;"
        exp = skbio.TreeNode.read([in_])
        obs = to_skbio_treenode(parse_newick(in_))
        self.assertEqual(obs.compare_rfd(exp), 0.0)

    def test_parse_newick_with_parens(self):
        # bug: parens are getting interpreted even if in quotes
        in_ = "(('foo(b)ar':1,baz:2)x:3)r;"
        exp = skbio.TreeNode.read([in_])
        obs = to_skbio_treenode(parse_newick(in_))
        self.assertEqual(obs.compare_rfd(exp), 0.0)

    def test_parse_newick(self):
        in_ = "((a:2,b):1,(c:4,d)y:20,e)r;"

        exp_bp = [1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0]
        exp_n = ['r', None, 'a', None, 'b', None, None, 'y', 'c', None, 'd',
                 None, None, 'e', None, None]
        exp_l = [0, 1, 2, 0, 0, 0, 0, 20, 4, 0, 0, 0, 0, 0, 0, 0]

        obs_bp = parse_newick(in_)

        npt.assert_equal(obs_bp.B, np.asarray(exp_bp, dtype=bool))

        for i, (e_n, e_l) in enumerate(zip(exp_n, exp_l)):
            self.assertEqual(obs_bp.name(i), e_n)
            self.assertEqual(obs_bp.length(i), e_l)

    def test_parse_newick_complex(self):
        in_ = "(((a:1,b:2.5)c:6,d:8,(e),(f,g,(h:1,i:2)j:1)k:1.2)l,m:2)r;"
        #         0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19
        exp_bp = [1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1,
                  1, 0, 1, 0, 0, 0, 0, 1, 0, 0]
        exp_n = ['r', 'l', 'c', 'a', None, 'b', None, None, 'd', None, None,
                 'e', None, None, 'k', 'f', None, 'g', None, 'j', 'h', None,
                 'i', None, None, None, None, 'm', None, None]
        exp_l = [0, 0, 6, 1, 0, 2.5, 0, 0, 8, 0, 0, 0, 0, 0, 1.2, 0, 0, 0, 0,
                 1, 1, 0, 2, 0, 0, 0, 0, 2, 0, 0]

        obs_bp = parse_newick(in_)

        npt.assert_equal(obs_bp.B, np.asarray(exp_bp, dtype=bool))

        for i, (e_n, e_l) in enumerate(zip(exp_n, exp_l)):
            self.assertEqual(obs_bp.name(i), e_n)
            self.assertEqual(obs_bp.length(i), e_l)

    def test_parse_newick_singledesc(self):
        in_ = "(((a)b)c,((d)e)f)r;"
        #         0  1  2  3  4  5  6  7  8  9 10 11 12 13
        exp_bp = [1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0]
        exp_n = ['r', 'c', 'b', 'a', None, None, None, 'f', 'e', 'd', None,
                 None, None, None]

        obs_bp = parse_newick(in_)

        npt.assert_equal(obs_bp.B, np.asarray(exp_bp, dtype=bool))

        for i, e_n in enumerate(exp_n):
            self.assertEqual(obs_bp.name(i), e_n)

    def test_parse_newick_unnamed_singledesc(self):
        in_ = "((a,b)c,d,(e))r;"

        exp_bp = [1, 1, 1, 0, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0]
        exp_n = ['r', 'c', 'a', None, 'b', None, None, 'd', None, None, 'e',
                 None, None, None]

        obs_bp = parse_newick(in_)

        npt.assert_equal(obs_bp.B, np.asarray(exp_bp, dtype=bool))

        for i, e_n in enumerate(exp_n):
            self.assertEqual(obs_bp.name(i), e_n)

    def test_parse_newick_name_with_semicolon(self):
        in_ = "((a,(b,c):5)'d','e; foo':10,((f))g)r;"

        exp_bp = [1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 0, 0]
        exp_n = ['r', 'd', 'a', None, None, 'b', None, 'c', None, None, None,
                 'e; foo', None, 'g', None, 'f', None, None, None, None]
        exp_l = [0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0]

        obs_bp = parse_newick(in_)

        npt.assert_equal(obs_bp.B, np.asarray(exp_bp, dtype=bool))

        for i, (e_n, e_l) in enumerate(zip(exp_n, exp_l)):
            self.assertEqual(obs_bp.name(i), e_n)
            self.assertEqual(obs_bp.length(i), e_l)


class JPlaceParseTests(TestCase):
    package = 'bp.tests'

    def setUp(self):
        self.jplacedata = open(self.get_data_path('200/placement.jplace'))
        self.jplacedata = self.jplacedata.read()
        no_edge_numbers = re.sub(r"{\d+}", '',
                                 json.loads(self.jplacedata)['tree'])
        self.tree = skbio.TreeNode.read([no_edge_numbers])

    def get_data_path(self, filename):
        # adapted from qiime2.plugin.testing.TestPluginBase
        return pkg_resources.resource_filename(self.package,
                                               'data/%s' % filename)

    def test_place_jplace_square_braces(self):
        self.jplacedata = json.loads(self.jplacedata)
        treestr = self.jplacedata['tree']
        treestr = re.sub(r"{(\d+)}", r"[\1]", treestr)
        self.jplacedata['tree'] = treestr
        self.jplacedata = json.dumps(self.jplacedata)

        exp_tree = self.tree
        obs_df, obs_tree = parse_jplace(self.jplacedata)
        obs_tree = to_skbio_treenode(obs_tree)
        self.assertEqual(obs_tree.compare_rfd(exp_tree), 0)
        for n in obs_tree.traverse(include_self=False):
            self.assertTrue(n.edge_num >= 0)

    def test_parse_jplace_simple(self):
        columns = ['fragment', 'edge_num', 'likelihood', 'like_weight_ratio',
                   'distal_length', 'pendant_length']
        exp_df = [["82", 361, 0.01013206496780672, 1, 0.02652932626620403,
                   0.039354548684623215],
                  ["99", 308, 0.04520741687623886, 1, 0.11020044356641526,
                   0.06550337922097477],
                  ["43", 309, 0.04054866161921744, 1, 0.010712923050783987,
                   0.020946988900520196],
                  ["195", 277, 0.01918907908397749, 1, 0.03065741838803451,
                   0.04513513498399864],
                  ["162", 55, 0.01758935282545493, 1, 0.0033199487685078776,
                   0.05388735804976052],
                  ["56", 81, 0.2366882303770561, 1, 0.04172580852519453,
                   0.0007060238727097983],
                  ["91", 105, 0.0001863393767883581, 1, 0.04578898721138839,
                   0.08655004339151215],
                  ["174", 89, 0.01216463967379211, 1, 0.04707020642820376,
                   0.045206727542450205],
                  ["5", 143, 0.012162345471765756, 1, 0.023797389484252734,
                   0.10447375403452556],
                  ["55", 139, 0.09563944060686769, 1, 0.014593217782258146,
                   0.04537214236560885]]
        exp_df = pd.DataFrame(exp_df, columns=columns)
        exp_tree = self.tree
        obs_df, obs_tree = parse_jplace(self.jplacedata)
        obs_tree = to_skbio_treenode(obs_tree)
        pdt.assert_frame_equal(obs_df, exp_df)
        self.assertEqual(obs_tree.compare_rfd(exp_tree), 0)

    def test_parse_jplace_multiple_per_fragment(self):
        columns = ['fragment', 'edge_num', 'likelihood', 'like_weight_ratio',
                   'distal_length', 'pendant_length']
        exp_df = [["82", 361, 0.01013206496780672, 1, 0.02652932626620403,
                   0.039354548684623215],
                  ["99", 308, 0.04520741687623886, 1, 0.11020044356641526,
                   0.06550337922097477],
                  # tied on like_weight_ratio but lower pendant
                  ["99", 309, 0.04520741687623886, 1, 0.11020044356641526,
                   0.00550337922097477],
                  ["55", 139, 0.09563944060686769, 1, 0.014593217782258146,
                   0.04537214236560885],
                  # tied higher like_weight_ratio
                  ["55", 138, 0.09563944060686769, 10, 0.014593217782258146,
                   0.04537214236560885]]
        exp_df = pd.DataFrame(exp_df, columns=columns)

        # ...adjust jplace data
        data = json.loads(self.jplacedata)
        keep = []
        for placement in data['placements']:
            if placement['n'][0] == '82':
                keep.append(placement)
            elif placement['n'][0] == '99':
                placement['p'].append([309, 0.04520741687623886, 1,
                                       0.11020044356641526,
                                       0.00550337922097477])
                keep.append(placement)
            elif placement['n'][0] == '55':
                placement['p'].append([138, 0.09563944060686769, 10,
                                       0.014593217782258146,
                                       0.04537214236560885])
                keep.append(placement)
        data['placements'] = keep
        data = json.dumps(data)
        exp_tree = self.tree
        obs_df, obs_tree = parse_jplace(data)
        obs_tree = to_skbio_treenode(obs_tree)
        pdt.assert_frame_equal(obs_df, exp_df)
        self.assertEqual(obs_tree.compare_rfd(exp_tree), 0)

    def test_parse_newick_linear_tree(self):
        # https://github.com/biocore/improved-octo-waddle/pull/48
        test = '((b:3)a:2)root:1;'

        # Test that we can parse these edge cases successfully, without
        # mistaking them for single-node trees
        topology = parse_newick(test)

        # Convert the tree to a skbio TreeNode to make checking easier
        skbio_tree = to_skbio_treenode(topology)
        self.assertEqual(skbio_tree.name, "root")
        self.assertEqual([n.name for n in skbio_tree.children], ["a"])
        self.assertEqual([n.name for n in skbio_tree.non_tips()], ["a"])
        self.assertEqual([n.name for n in skbio_tree.tips()], ["b"])


if __name__ == '__main__':
    main()
